@@ -6,13 +6,18 @@
 ## License: GPL Version 2
 ## Copyright: TGRMN Software and contributors
 
-from Utils import getTreeFromXml
+from __future__ import absolute_import, print_function
+
+import sys
+from .Utils import getTreeFromXml, deunicodise, encode_to_s3, decode_from_s3
 
 try:
     import xml.etree.ElementTree as ET
 except ImportError:
     import elementtree.ElementTree as ET
 
+PY3 = (sys.version_info >= (3,0))
+
 class Grantee(object):
     ALL_USERS_URI = "http://acs.amazonaws.com/groups/global/AllUsers"
     LOG_DELIVERY_URI = "http://acs.amazonaws.com/groups/s3/LogDelivery"
@@ -25,11 +30,11 @@ class Grantee(object):
         self.permission = None
 
     def __repr__(self):
-        return 'Grantee("%(tag)s", "%(name)s", "%(permission)s")' % {
+        return repr('Grantee("%(tag)s", "%(name)s", "%(permission)s")' % {
             "tag" : self.tag,
             "name" : self.name,
             "permission" : self.permission
-        }
+        })
 
     def isAllUsers(self):
         return self.tag == "URI" and self.name == Grantee.ALL_USERS_URI
@@ -69,7 +74,7 @@ class GranteeLogDelivery(Grantee):
         self.permission = permission
 
 class ACL(object):
-    EMPTY_ACL = "<AccessControlPolicy><Owner><ID></ID></Owner><AccessControlList></AccessControlList></AccessControlPolicy>"
+    EMPTY_ACL = b"<AccessControlPolicy><Owner><ID></ID></Owner><AccessControlList></AccessControlList></AccessControlPolicy>"
 
     def __init__(self, xml = None):
         if not xml:
@@ -79,7 +84,7 @@ class ACL(object):
         self.owner_id = ""
         self.owner_nick = ""
 
-        tree = getTreeFromXml(xml)
+        tree = getTreeFromXml(encode_to_s3(xml))
         self.parseOwner(tree)
         self.parseGrants(tree)
 
@@ -184,10 +189,9 @@ class ACL(object):
             self.grantees = [g for g in self.grantees if not (g.name.lower() == name or g.display_name.lower() == name)]
         else:
             self.grantees = [g for g in self.grantees if not ((g.display_name.lower() == name and g.permission.upper() == permission)\
-				 or (g.name.lower() == name and g.permission.upper() ==  permission))]
+                or (g.name.lower() == name and g.permission.upper() ==  permission))]
 
-
-    def __str__(self):
+    def get_printable_tree(self):
         tree = getTreeFromXml(ACL.EMPTY_ACL)
         tree.attrib['xmlns'] = "http://s3.amazonaws.com/doc/2006-03-01/"
         owner = tree.find(".//Owner//ID")
@@ -195,10 +199,21 @@ class ACL(object):
         acl = tree.find(".//AccessControlList")
         for grantee in self.grantees:
             acl.append(grantee.getElement())
-        return ET.tostring(tree)
+        return tree
+
+    def __unicode__(self):
+        return decode_from_s3(ET.tostring(self.get_printable_tree()))
+
+    def __str__(self):
+        if PY3:
+            # Return unicode
+            return ET.tostring(self.get_printable_tree(), encoding="unicode")
+        else:
+            # Return bytes
+            return ET.tostring(self.get_printable_tree())
 
 if __name__ == "__main__":
-    xml = """<?xml version="1.0" encoding="UTF-8"?>
+    xml = b"""<?xml version="1.0" encoding="UTF-8"?>
 <AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
 <Owner>
     <ID>12345678901234567890</ID>
@@ -222,11 +237,11 @@ if __name__ == "__main__":
 </AccessControlPolicy>
     """
     acl = ACL(xml)
-    print "Grants:", acl.getGrantList()
+    print("Grants:", acl.getGrantList())
     acl.revokeAnonRead()
-    print "Grants:", acl.getGrantList()
+    print("Grants:", acl.getGrantList())
     acl.grantAnonRead()
-    print "Grants:", acl.getGrantList()
-    print acl
+    print("Grants:", acl.getGrantList())
+    print(acl)
 
 # vim:et:ts=4:sts=4:ai
