@@ -1,12 +1,15 @@
+# -*- coding: utf-8 -*-
+
 ## Amazon CloudFront support
 ## Author: Michal Ludvig <michal@logix.cz>
 ##         http://www.logix.cz/michal
 ## License: GPL Version 2
 ## Copyright: TGRMN Software and contributors
 
+from __future__ import absolute_import
+
 import sys
 import time
-import httplib
 import random
 from datetime import datetime
 from logging import debug, info, warning, error
@@ -16,14 +19,14 @@ try:
 except ImportError:
     import elementtree.ElementTree as ET
 
-from S3 import S3
-from Config import Config
-from Exceptions import *
-from Utils import getTreeFromXml, appendXmlTextNode, getDictFromTree, dateS3toPython, getBucketFromHostname, getHostnameFromBucket
-from Crypto import sign_string_v2
-from S3Uri import S3Uri, S3UriS3
-from FileLists import fetch_remote_list
-from ConnMan import ConnMan
+from .S3 import S3
+from .Config import Config
+from .Exceptions import *
+from .Utils import getTreeFromXml, appendXmlTextNode, getDictFromTree, dateS3toPython, getBucketFromHostname, getHostnameFromBucket, deunicodise, urlencode_string, convertHeaderTupleListToDict
+from .Crypto import sign_string_v2
+from .S3Uri import S3Uri, S3UriS3
+from .ConnMan import ConnMan
+from .SortedDict import SortedDict
 
 cloudfront_api_version = "2010-11-01"
 cloudfront_resource = "/%(api_ver)s/distribution" % { 'api_ver' : cloudfront_api_version }
@@ -61,11 +64,11 @@ class DistributionSummary(object):
     def parse(self, tree):
         self.info = getDictFromTree(tree)
         self.info['Enabled'] = (self.info['Enabled'].lower() == "true")
-        if self.info.has_key("CNAME") and type(self.info['CNAME']) != list:
+        if "CNAME" in self.info and type(self.info['CNAME']) != list:
             self.info['CNAME'] = [self.info['CNAME']]
 
     def uri(self):
-        return S3Uri("cf://%s" % self.info['Id'])
+        return S3Uri(u"cf://%s" % self.info['Id'])
 
 class DistributionList(object):
     ## Example:
@@ -121,7 +124,7 @@ class Distribution(object):
         self.info['DistributionConfig'] = DistributionConfig(tree = tree.find(".//DistributionConfig"))
 
     def uri(self):
-        return S3Uri("cf://%s" % self.info['Id'])
+        return S3Uri(u"cf://%s" % self.info['Id'])
 
 class DistributionConfig(object):
     ## Example:
@@ -153,14 +156,14 @@ class DistributionConfig(object):
     def parse(self, tree):
         self.info = getDictFromTree(tree)
         self.info['Enabled'] = (self.info['Enabled'].lower() == "true")
-        if not self.info.has_key("CNAME"):
+        if "CNAME" not in self.info:
             self.info['CNAME'] = []
         if type(self.info['CNAME']) != list:
             self.info['CNAME'] = [self.info['CNAME']]
         self.info['CNAME'] = [cname.lower() for cname in self.info['CNAME']]
-        if not self.info.has_key("Comment"):
+        if "Comment" not in self.info:
             self.info['Comment'] = ""
-        if not self.info.has_key("DefaultRootObject"):
+        if "DefaultRootObject" not in self.info:
             self.info['DefaultRootObject'] = ""
         ## Figure out logging - complex node not parsed by getDictFromTree()
         logging_nodes = tree.findall(".//Logging")
@@ -169,7 +172,7 @@ class DistributionConfig(object):
             logging_dict['Bucket'], success = getBucketFromHostname(logging_dict['Bucket'])
             if not success:
                 warning("Logging to unparsable bucket name: %s" % logging_dict['Bucket'])
-            self.info['Logging'] = S3UriS3("s3://%(Bucket)s/%(Prefix)s" % logging_dict)
+            self.info['Logging'] = S3UriS3(u"s3://%(Bucket)s/%(Prefix)s" % logging_dict)
         else:
             self.info['Logging'] = None
 
@@ -284,12 +287,11 @@ class InvalidationBatch(object):
 
     def __str__(self):
         tree = ET.Element("InvalidationBatch")
-        s3 = S3(Config())
 
         for path in self.paths:
             if len(path) < 1 or path[0] != "/":
                 path = "/" + path
-            appendXmlTextNode("Path", s3.urlencode_string(path), tree)
+            appendXmlTextNode("Path", urlencode_string(path), tree)
         appendXmlTextNode("CallerReference", self.reference, tree)
         return ET.tostring(tree)
 
@@ -396,7 +398,7 @@ class CloudFront(object):
                     break
                 warning("Still waiting...")
                 time.sleep(10)
-        headers = {}
+        headers = SortedDict(ignore_case = True)
         headers['if-match'] = response['headers']['etag']
         response = self.send_request("DeleteDist", dist_id = cfuri.dist_id(),
                                      headers = headers)
@@ -423,7 +425,7 @@ class CloudFront(object):
         debug("SetDistConfig(): Etag = %s" % etag)
         request_body = str(dist_config)
         debug("SetDistConfig(): request_body: %s" % request_body)
-        headers = {}
+        headers = SortedDict(ignore_case = True)
         headers['if-match'] = etag
         response = self.send_request("SetDistConfig", dist_id = cfuri.dist_id(),
                                      body = request_body, headers = headers)
@@ -447,28 +449,33 @@ class CloudFront(object):
             paths = new_paths
 
         # uri could be either cf:// or s3:// uri
-        cfuri = self.get_dist_name_for_bucket(uri)
+        cfuris = self.get_dist_name_for_bucket(uri)
         if len(paths) > 999:
             try:
                 tmp_filename = Utils.mktmpfile()
-                f = open(tmp_filename, "w")
-                f.write("\n".join(paths)+"\n")
-                f.close()
+                with open(deunicodise(tmp_filename), "w") as fp:
+                    fp.write(deunicodise("\n".join(paths)+"\n"))
                 warning("Request to invalidate %d paths (max 999 supported)" % len(paths))
                 warning("All the paths are now saved in: %s" % tmp_filename)
             except:
                 pass
             raise ParameterError("Too many paths to invalidate")
-        invalbatch = InvalidationBatch(distribution = cfuri.dist_id(), paths = paths)
-        debug("InvalidateObjects(): request_body: %s" % invalbatch)
-        response = self.send_request("Invalidate", dist_id = cfuri.dist_id(),
-                                     body = str(invalbatch))
-        response['dist_id'] = cfuri.dist_id()
-        if response['status'] == 201:
-            inval_info = Invalidation(response['data']).info
-            response['request_id'] = inval_info['Id']
-        debug("InvalidateObjects(): response: %s" % response)
-        return response
+
+        responses = []
+        for cfuri in cfuris:
+            invalbatch = InvalidationBatch(distribution = cfuri.dist_id(), paths = paths)
+            debug("InvalidateObjects(): request_body: %s" % invalbatch)
+            response = self.send_request("Invalidate", dist_id = cfuri.dist_id(),
+                                         body = str(invalbatch))
+            response['dist_id'] = cfuri.dist_id()
+            if response['status'] == 201:
+                inval_info = Invalidation(response['data']).info
+                response['request_id'] = inval_info['Id']
+            debug("InvalidateObjects(): response: %s" % response)
+
+            responses.append(response)
+
+        return responses
 
     def GetInvalList(self, cfuri):
         if cfuri.type != "cf":
@@ -490,7 +497,9 @@ class CloudFront(object):
     ## Low-level methods for handling CloudFront requests
     ## --------------------------------------------------
 
-    def send_request(self, op_name, dist_id = None, request_id = None, body = None, headers = {}, retries = _max_retries):
+    def send_request(self, op_name, dist_id = None, request_id = None, body = None, headers = None, retries = _max_retries):
+        if headers is None:
+            headers = SortedDict(ignore_case = True)
         operation = self.operations[op_name]
         if body:
             headers['content-type'] = 'text/plain'
@@ -502,7 +511,7 @@ class CloudFront(object):
         response = {}
         response["status"] = http_response.status
         response["reason"] = http_response.reason
-        response["headers"] = dict(http_response.getheaders())
+        response["headers"] = convertHeaderTupleListToDict(http_response.getheaders())
         response["data"] =  http_response.read()
         ConnMan.put(conn)
 
@@ -529,14 +538,14 @@ class CloudFront(object):
                    operation['resource'] % { 'dist_id' : dist_id, 'request_id' : request_id })
 
         if not headers:
-            headers = {}
+            headers = SortedDict(ignore_case = True)
 
-        if headers.has_key("date"):
-            if not headers.has_key("x-amz-date"):
+        if "date" in headers:
+            if "x-amz-date" not in headers:
                 headers["x-amz-date"] = headers["date"]
             del(headers["date"])
 
-        if not headers.has_key("x-amz-date"):
+        if "x-amz-date" not in headers:
             headers["x-amz-date"] = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())
 
         if len(self.config.access_token)>0:
@@ -568,9 +577,9 @@ class CloudFront(object):
         return (self._max_retries - retries + 1) * 3
 
     def get_dist_name_for_bucket(self, uri):
-        if (uri.type == "cf"):
-            return uri
-        if (uri.type != "s3"):
+        if uri.type == "cf":
+            return [uri]
+        if uri.type != "s3":
             raise ParameterError("CloudFront or S3 URI required instead of: %s" % uri)
 
         debug("_get_dist_name_for_bucket(%r)" % uri)
@@ -578,23 +587,30 @@ class CloudFront(object):
             response = self.GetList()
             CloudFront.dist_list = {}
             for d in response['dist_list'].dist_summs:
-                if d.info.has_key("S3Origin"):
-                    CloudFront.dist_list[getBucketFromHostname(d.info['S3Origin']['DNSName'])[0]] = d.uri()
-                elif d.info.has_key("CustomOrigin"):
+                distListIndex = ""
+
+                if "S3Origin" in d.info:
+                    distListIndex = getBucketFromHostname(d.info['S3Origin']['DNSName'])[0]
+                elif "CustomOrigin" in d.info:
                     # Aral: This used to skip over distributions with CustomOrigin, however, we mustn't
                     #       do this since S3 buckets that are set up as websites use custom origins.
                     #       Thankfully, the custom origin URLs they use start with the URL of the
                     #       S3 bucket. Here, we make use this naming convention to support this use case.
                     distListIndex = getBucketFromHostname(d.info['CustomOrigin']['DNSName'])[0];
                     distListIndex = distListIndex[:len(uri.bucket())]
-                    CloudFront.dist_list[distListIndex] = d.uri()
                 else:
                     # Aral: I'm not sure when this condition will be reached, but keeping it in there.
                     continue
+
+                if CloudFront.dist_list.get(distListIndex, None) is None:
+                    CloudFront.dist_list[distListIndex] = set() 
+
+                CloudFront.dist_list[distListIndex].add(d.uri())
+
             debug("dist_list: %s" % CloudFront.dist_list)
         try:
             return CloudFront.dist_list[uri.bucket()]
-        except Exception, e:
+        except Exception as e:
             debug(e)
             raise ParameterError("Unable to translate S3 URI to CloudFront distribution name: %s" % uri)
 
@@ -624,8 +640,8 @@ class Cmd(object):
         cf = CloudFront(Config())
         cfuris = []
         for arg in args:
-            uri = cf.get_dist_name_for_bucket(S3Uri(arg))
-            cfuris.append(uri)
+            uris = cf.get_dist_name_for_bucket(S3Uri(arg))
+            cfuris.extend(uris)
         return cfuris
 
     @staticmethod
@@ -634,16 +650,16 @@ class Cmd(object):
         if not args:
             response = cf.GetList()
             for d in response['dist_list'].dist_summs:
-                if d.info.has_key("S3Origin"):
+                if "S3Origin" in d.info:
                     origin = S3UriS3.httpurl_to_s3uri(d.info['S3Origin']['DNSName'])
-                elif d.info.has_key("CustomOrigin"):
+                elif "CustomOrigin" in d.info:
                     origin = "http://%s/" % d.info['CustomOrigin']['DNSName']
                 else:
                     origin = "<unknown>"
                 pretty_output("Origin", origin)
                 pretty_output("DistId", d.uri())
                 pretty_output("DomainName", d.info['DomainName'])
-                if d.info.has_key("CNAME"):
+                if "CNAME" in d.info:
                     pretty_output("CNAMEs", ", ".join(d.info['CNAME']))
                 pretty_output("Status", d.info['Status'])
                 pretty_output("Enabled", d.info['Enabled'])
@@ -654,16 +670,16 @@ class Cmd(object):
                 response = cf.GetDistInfo(cfuri)
                 d = response['distribution']
                 dc = d.info['DistributionConfig']
-                if dc.info.has_key("S3Origin"):
+                if "S3Origin" in dc.info:
                     origin = S3UriS3.httpurl_to_s3uri(dc.info['S3Origin']['DNSName'])
-                elif dc.info.has_key("CustomOrigin"):
+                elif "CustomOrigin" in dc.info:
                     origin = "http://%s/" % dc.info['CustomOrigin']['DNSName']
                 else:
                     origin = "<unknown>"
                 pretty_output("Origin", origin)
                 pretty_output("DistId", d.uri())
                 pretty_output("DomainName", d.info['DomainName'])
-                if dc.info.has_key("CNAME"):
+                if "CNAME" in dc.info:
                     pretty_output("CNAMEs", ", ".join(dc.info['CNAME']))
                 pretty_output("Status", d.info['Status'])
                 pretty_output("Comment", dc.info['Comment'])
@@ -723,7 +739,7 @@ class Cmd(object):
             raise ParameterError("Too many parameters. Modify one Distribution at a time.")
         try:
             cfuri = Cmd._parse_args(args)[0]
-        except IndexError, e:
+        except IndexError:
             raise ParameterError("No valid Distribution URI found.")
         response = cf.ModifyDistribution(cfuri,
                                          cnames_add = Cmd.options.cf_cnames_add,
